from abc import ABC, abstractmethod
from typing import List

from .distance_function import DistanceFunction
from .vector_distance import VectorDistance
from .vector import Vector


class VectorDB(ABC):

    @abstractmethod
    def insert_vector(self, vector: Vector) -> Vector:
        ...

    @abstractmethod
    def read_vector(self, id: str) -> Vector | None:
        ...

    @abstractmethod
    def update_vector(self, vector: Vector) -> Vector:
        ...

    @abstractmethod
    def delete_vector(self, id: str) -> None:
        ...

    @abstractmethod
    def get_neighbor_vectors(self, vector: Vector, n: int) -> List[VectorDistance]:
        ...



from enum import Enum


class DistanceFunction(Enum):
    """Distance function to measure how var the vector to surrounding vector in vector db"""
    COSINE = "cosine"
    HAMMING = "hamming"
    JACCARD = "jaccard"
    L1 = "l1"
    L2 = "l2"
    MAX_INNER_PRODUCT = "max_inner_product"
    EUCLIDEAN = "euclid"
    DOT = "dot"
    MANHATTAN = "manhattan"



import logging
from qdrant_client import QdrantClient
from qdrant_client.models import VectorParams, Distance, PointStruct, ScoredPoint

from typing import List

from pyvectordb.distance_function import DistanceFunction
from pyvectordb.driver import VectorDB
from pyvectordb.vector_distance import VectorDistance
from pyvectordb.vector import Vector


class QdrantDB(VectorDB):
    
    def __init__(
        self, 
        host: str, 
        api_key: str, 
        port: int, 
        collection: str,
        vector_size: int,
        distance_function: DistanceFunction=DistanceFunction.EUCLIDEAN,
    ) -> None:
        super().__init__()
        
        self.host = host or self.__raise_value_error("host")
        self.api_key = api_key or self.__raise_value_error("api_key")
        self.port = port or self.__raise_value_error("port")
        self.collection = collection or self.__raise_value_error("collection")
        self.vector_size = vector_size or self.__raise_value_error("vector_size")
        self.distance_function = distance_function or self.__raise_value_error("distance_function")
        
        self.client: QdrantClient = None
        
        self.__init__client()
        self.__init_collection()
    
    @staticmethod
    def __raise_value_error(param: str):
        raise ValueError(f"{param} is required")
    
    def __init__client(self) -> None:
        if self.client is None:
            self.client = QdrantClient(host=self.host, port=self.port, api_key=self.api_key, https=False)
    
    def __init_collection(self) -> None:
        if not self.client.collection_exists(self.collection):
            self.client.create_collection(
                collection_name=self.collection,
                vectors_config=VectorParams(
                    size=self.vector_size, 
                    distance=self.__get_distance_function(self.distance_function)
                ),
            )
            
    def __get_distance_function(self, distance: DistanceFunction) -> Distance:
        if distance == DistanceFunction.COSINE: 
            distance_func = Distance.COSINE
        elif distance == DistanceFunction.EUCLIDEAN: 
            distance_func = Distance.EUCLID
        elif distance == DistanceFunction.DOT: 
            distance_func = Distance.DOT
        elif distance == DistanceFunction.MANHATTAN: 
            distance_func = Distance.MANHATTAN
        else:
            raise ValueError("distance function unavailable on qdrant")
        
        return distance_func

    def insert_vector(self, vector: Vector) -> Vector:
        vector_id = vector.get_id()

        self.client.upsert(
            collection_name=self.collection,
            points=[
                PointStruct(
                    id=vector_id,
                    vector=vector.embedding,
                    payload={"metadata": vector.metadata}
                )
            ],
            wait=False
        )
        
        return Vector(
            embedding=vector.embedding,
            vector_id=vector_id,
            metadata=vector.metadata,
        )

    def read_vector(self, id: int) -> Vector | None:
        records = self.client.retrieve(
            collection_name=self.collection,
            ids=[id],
            with_payload=True,
            with_vectors=True,
        )
        if len(records) == 0:
            return None
        
        record = records[0]
        return Vector(
            embedding=record.vector,
            vector_id=record.id,
            metadata=record.payload.get("metadata"),
        )

    def update_vector(self, vector: Vector) -> Vector:
        if self.read_vector(vector.get_id()) is None:
            raise ValueError("vector not exist")
        
        self.client.upsert(
            collection_name=self.collection,
            points=[
                PointStruct(
                    id=vector.get_id(),
                    vector=vector.embedding,
                    payload={"metadata": vector.metadata}
                )
            ],
            wait=False
        )
        return Vector(
            embedding=vector.embedding,
            vector_id=vector.get_id(),
            metadata=vector.metadata,
        )

    def delete_vector(self, id: int) -> None:
        self.client.delete(
            collection_name=self.collection,
            points_selector=[id],
            wait=False
        )

    def get_neighbor_vectors(
        self,
        vector: Vector,
        n: int,
    ) -> List[VectorDistance]:    
        scored_points: List[ScoredPoint] = self.client.search(
            collection_name=self.collection,
            query_vector=vector.embedding,
            with_payload=True,
            with_vectors=True,
            limit=n,
        )
        vector_distances  = []
        for point in scored_points:
            vector_distance = VectorDistance(
                vector=Vector(
                    embedding=point.vector,
                    vector_id=point.id,
                    metadata=point.payload.get("metadata"),
                ),
                distance=point.score,
            )
            vector_distances.append(vector_distance)
        return vector_distances



from datetime import datetime
from pgvector.sqlalchemy import Vector
from sqlalchemy.orm import DeclarativeBase, mapped_column, Mapped
from sqlalchemy import Column, String, DateTime, Table


class Base(DeclarativeBase):
    __abstract__ = True
    def to_dict(self):
        return {field.name:getattr(self, field.name) for field in self.__table__.c}


def get_vector_orm(tablename: str) -> Base:
        
    class VectorORM(Base):
        __tablename__ = tablename
        
        id: Mapped[str] = Column(String, primary_key=True)
        embedding: Mapped[Vector] = mapped_column(Vector())
        metadata_: Mapped[str] = Column(String, nullable=False, name="metadata")
        created_at: Mapped[datetime] = Column(DateTime, nullable=False, default=datetime.now())
        
    return VectorORM


# def get_vector_orm(table_name: str, engine) -> DefaultVectorORM:
#     v = DefaultVectorORM
#     v.__tablename__ = table_name
    
#     v.__table__ = Table(table_name, Base.metadata, autoload_with=engine)
#     return v


import logging
import chromadb
from chromadb.config import Settings
from typing import List

from pyvectordb.distance_function import DistanceFunction
from pyvectordb.vector_distance import VectorDistance
from pyvectordb.vector import Vector
from pyvectordb.driver import VectorDB

from .distance import Distance


class ChromaDB(VectorDB):
    
    def __init__(
        self, 
        host: str, 
        port: str,
        auth_provider: str,
        auth_credentials: str,
        collection_name: str,
        distance_function: DistanceFunction=DistanceFunction.L2,
    ) -> None:
        super().__init__()
        
        self.host = host or self.__raise_value_error("host")
        self.port = port or self.__raise_value_error("port")
        self.auth_provider = auth_provider or self.__raise_value_error("auth_provider")
        self.auth_credentials = auth_credentials or self.__raise_value_error("auth_credentials")
        self.collection_name = collection_name or self.__raise_value_error("collection_name")
        self.distance_function = distance_function or self.__raise_value_error("distance_function")
        
        self.client = None
        self.collection = None
        
        self.__init_client()
        self.__health_check()
        self.__init_collection()
    
    @staticmethod
    def __raise_value_error(param: str):
        raise ValueError(f"{param} is required")
    
    def __init_client(self) -> None:
        if self.client is None:
            self.client = chromadb.HttpClient(
                host=self.host,
                port=self.port,
                ssl=False,
                headers={"X-Chroma-Token": self.auth_credentials},
                settings=Settings(
                    chroma_client_auth_provider=self.auth_provider,
                    chroma_client_auth_credentials=self.auth_credentials,
                )
            )
    
    def __health_check(self) -> None:
        self.client.heartbeat()
        
    def __init_collection(self) -> None:
        if self.collection == None:
            distance_func = self.__get_distance_function(self.distance_function)
            self.collection = self.client.get_or_create_collection(
                name=self.collection_name,
                metadata={"hnsw:space": distance_func.value}
            )

    def insert_vector(self, vector: Vector) -> Vector:
        self.collection.add(
            ids=[vector.get_id()],
            embeddings=[vector.embedding],
            metadatas=[vector.metadata],
        )
        return vector

    def read_vector(self, id: str) -> Vector | None:
        result: dict = self.collection.get(
            ids=id,
            include=["metadatas", "documents", "embeddings"]
        )
        return Vector(
            vector_id=result.get("ids")[0],
            embedding=result.get("embeddings")[0],
            metadata=result.get("metadatas")[0].get("metadata")
        )

    def update_vector(self, vector: Vector) -> Vector:
        self.collection.update(
            ids=[vector.get_id()],
            embeddings=[vector.embedding],
            metadatas=[{"metadata": vector.metadata}],
        )
        return vector

    def delete_vector(self, id: str) -> None:
        self.collection.delete(
            ids=[id],
        )

    def get_neighbor_vectors(
        self, 
        vector: Vector, 
        n: int, 
    ) -> List[VectorDistance]:
        result: dict = self.collection.query(
            query_embeddings=[vector.embedding],
            n_results=n,
            include = ["metadatas", "documents", "distances", "embeddings"],
        )
        
        vds = []
        for i, _ in enumerate(result.get("ids")[0]):
            vd = VectorDistance(
                vector=Vector(
                    vector_id=result.get("ids")[0][i],
                    embedding=result.get("embeddings")[0][i],
                    metadata=result.get("metadatas")[0][i]
                ),
                distance=result.get("distances")[0][i]
            )
            vds.append(vd)
                              
        return vds

    def __get_distance_function(self, distance: DistanceFunction) -> Distance:
        if distance == DistanceFunction.L2:
            return Distance.SQUARED_L2
        elif distance == DistanceFunction.MAX_INNER_PRODUCT:
            return Distance.INNER_PRODUCT
        elif distance == DistanceFunction.COSINE:
            return Distance.COSINE_SIMILARITY
        else:
            raise ValueError(f"distance function unavailable on chromadb. [{Distance}]")


from .vector import Vector


class VectorDistance:
    
    def __init__(self, vector: Vector, distance: float) -> None:
        self.vector: Vector=vector
        self.distance = distance
        
    def __str__(self) -> str:
        return f"{self.vector}, distance={self.distance}"
    
    def __repr__(self) -> str:
        return self.__str__()


from sqlalchemy import create_engine, select, text
from sqlalchemy.orm import sessionmaker, Session
from typing import Any, Generator, List

from pyvectordb.distance_function import DistanceFunction
from pyvectordb.vector_distance import VectorDistance
from pyvectordb.driver import VectorDB
from pyvectordb.vector import Vector

from .model import get_vector_orm


class PgvectorDB(VectorDB):
    
    def __init__(
        self,
        db_user: str,
        db_password: str,
        db_host: str,
        db_port: str,
        db_name: str,
        collection: str,
        distance_function: DistanceFunction=DistanceFunction.L2,
    ) -> None:
        super().__init__()
        
        self.db_user = db_user or self.__raise_value_error("db_user")
        self.db_password = db_password or self.__raise_value_error("db_password")
        self.db_host = db_host or self.__raise_value_error("db_host")
        self.db_port = db_port or self.__raise_value_error("db_port")
        self.db_name = db_name or self.__raise_value_error("db_name")  
        self.collection = collection or self.__raise_value_error("collection")  
        self.distance_function = distance_function or self.__raise_value_error("distance_function")
        
        self.__engine = None
        
        self.__init_engine()
        self.conn = next(self.__get_db_session())
        
        self.__init_collection()
        self.__vector_orm = get_vector_orm(self.collection)

    @staticmethod
    def __raise_value_error(param: str):
        raise ValueError(f"{param} is required")
    
    def __init_engine(self) -> None:
        if self.__engine == None:
            self.__engine = create_engine(
                f'postgresql://{self.db_user}:{self.db_password}@{self.db_host}:{self.db_port}/{self.db_name}', 
                echo=True,
                connect_args={}
            )
    
    def __get_db_session(self) -> Generator[Session, None, None]:
        SessionLocal = sessionmaker(
            autocommit=False, 
            autoflush=False, 
            bind=self.__engine
        )

        db = SessionLocal()
        try:
            yield db
        finally:
            db.close()
            
    def __init_collection(self) -> None:
        query = f"""
CREATE TABLE IF NOT EXISTS {self.collection} (
    id text PRIMARY KEY,
    embedding vector,
    metadata text,
    created_at timestamptz DEFAULT now()
);
"""
        self.conn.execute(text(query))
        self.conn.commit()
        

    def insert_vector(self, vector: Vector) -> Vector:
        conn = self.conn
        
        v = self.__vector_orm(
            id=vector.get_id(),
            embedding=vector.embedding,
            metadata_=vector.metadata,
        )
        
        conn.add(v)
        conn.commit()
        conn.refresh(v)

        return Vector(
            embedding=v.embedding,
            vector_id=v.id,
            metadata=v.metadata_,
        )
        
    def read_vector(self, id: int) -> Vector | None:
        v = self.__read_vector_orm(id)
        
        if v is None: 
            return None
        
        return Vector(
            embedding=v.embedding,
            vector_id=v.id,
            metadata=v.metadata_,
        )   
    
    def update_vector(self, vector: Vector) -> Vector:
        conn = self.conn
        
        v = self.__read_vector_orm(vector.id)
        if v is None: raise ValueError("vector not found in database")
        
        v.embedding = vector.embedding
        v._metadata = vector.metadata
        
        conn.add(v)
        conn.commit()
        
        return Vector(
            embedding=v.embedding,
            vector_id=v.id,
            metadata=v.metadata_,
        )

    def delete_vector(self, id: int) -> None:
        conn = self.conn
        v = self.__read_vector_orm(id)
        conn.delete(v)
        conn.commit()
    
    def get_neighbor_vectors(
        self, 
        vector: Vector, 
        n: int=5, 
    ) -> List[VectorDistance]:
        vectordistances = []
        distance_func = self.__get_distance_function(self.distance_function)

        q = self.conn.execute(
            select(
                self.__vector_orm,
                distance_func(vector.embedding).label("distance")
            )
            .order_by(distance_func(vector.embedding))
            .limit(n)
        )
        results = q.all()
        
        for r in results:
            vector = Vector(
                embedding=r[0].embedding,
                vector_id=r[0].id,
                metadata=r[0].metadata_,
            )
            distance = r[1]
            vectordistances.append(
                VectorDistance(vector, distance)
            )
        return vectordistances
    
    def __get_distance_function(self, selected_distance_function: DistanceFunction) -> Any:
        if selected_distance_function==DistanceFunction.L2:
            return self.__vector_orm.embedding.l2_distance
        elif selected_distance_function==DistanceFunction.MAX_INNER_PRODUCT:
            return self.__vector_orm.embedding.max_inner_product
        elif selected_distance_function==DistanceFunction.COSINE:
            return self.__vector_orm.embedding.cosine_distance
        elif selected_distance_function==DistanceFunction.L1:
            return self.__vector_orm.embedding.l1_distance
        elif selected_distance_function==DistanceFunction.HAMMING:
            return self.__vector_orm.embedding.hamming_distance
        elif selected_distance_function==DistanceFunction.JACCARD:
            return self.__vector_orm.embedding.jaccard_distance
        else:
            raise ValueError("distance function unavailable on pgvector")
        
    def __read_vector_orm(self, id: int) -> object | None:
        v = self.conn.execute(
            select(self.__vector_orm)
            .where(self.__vector_orm.id == id)
        ).one_or_none()
        if v is None:
            return v
        
        return v[0]